pub struct SlotSequentialStream { /* private fields */ }Expand description
A streams that yeild FumaroleEvent in slot order, meaning that while a slot is active, only events from that
slot will be yielded, and once the slot ends, events from the next slot will be yielded, and so on. So while a slot
has not ended, event’s from that slot won’t be interleaved with events from other slots.
This is usefulfor consumers that want to process events in slot order and don’t care about processing events from multiple slots concurrently.
If fumarole download two slots in parallel, (when replaying from the past), say the slot 1, 2. Then whatever slot gives yield the first event, say slot 2, will be the only slot that is yielded until it ends, and then events from slot 1 will be yielded.
Implementations
Sourceimpl SlotSequentialStream
impl SlotSequentialStream
Sourcepub fn commit(&mut self)
pub fn commit(&mut self)
See FumaroleStream::commit for more details.